// Main file for unit testing with catch. The directive below tells cmake to
// generate a main method for the tests. Define tests in a separate cpp file
// with only the include below. Do not define CATCH_CONFIG_MAIN anywhere else.
// See correct-print-options in TestSetup.cpp.
#define CATCH_CONFIG_MAIN
#include "catch.hpp"
